const constants = require("./constants");

/** 
 * A custom error for operational (or expected) error. 
 * Operational errors will get logged to papertrial, but not to
 * slack.
 * */
class OperationalError extends Error {

    /**
     * Create an operational error.
     * @param {string} message - The error message
     * @param {string} [data] - Additional contextual data
     * */
    constructor(message, data = "") {
        super(message);
        Error.captureStackTrace(this, OperationalError);
        this.name = "OperationalError";
        this.isOperational = true;
        this.data = { data: data, stack: this.stack };
    }
}

/** 
 * A custom error for Non operational errors (or unexpected) errors. 
 * NonOperationalError will get sent to slack, and papertrial, and are errors 
 * that need to be addressed by the team. 
 *
 * Error will by default be sent to "alerts_web", but you can pass in a slack channel if you 
 * need the error to go somewhere else. Or you can set process.env.NON_OPERATIONAL_SLACK_CHANNEL 
 * to override the default slack channel. If you pass in a slack channel to the constructor, it will
 * still use that channel, even if the env variable is set.
 * */
class NonOperationalError extends Error {

    /** 
     * Create a non operational error 
     * @param {string} message - The error message 
     * @param {string} [data] - Additional contextual data 
     * @param {...import("./constants").slackChannels} [slackChannel] - The slack channel to send the error to, defaults to "alerts_web".
     * */
    constructor(message, data = "", slackChannel = null) {
        super(message);
        Error.captureStackTrace(this, NonOperationalError);
        this.name = "NonOperationalError";
        this.isOperational = false;
        this.data = { data: data, stack: this.stack };
        this.slackChannel = slackChannel;
    }
}

/** 
 * Custom sub error type of OperationalError that 
 * also takes in a status code. Use this error if you want 
 * the message logged to papertrial, sent to the client, and a status code other 
 * than 500 to be returned. 
 *
 * Be mindful when using this error type, as it will have its error message returned
 * to the client. We want to be careful what information we expose.
 * */
class HttpError extends OperationalError {

    /** 
     * Create an http error, with message and status code. 
     * @param {string} message - The error message 
     * @param {number} statusCode - The http status code 
     * @param {string} [data] - Additional contextual data 
     * */
    constructor(message, statusCode, data = "") {
        super(message);
        Error.captureStackTrace(this, HttpError);
        this.name = "HttpError";
        this.statusCode = statusCode;
        this.data = { data: data, stack: this.stack };
    }
}

module.exports = {
    OperationalError,
    NonOperationalError,
    HttpError
}
