jest.mock("@firstfleet/fflogger", () => {
    const EventEmitter = require("events")
    const emitter = new EventEmitter();

    return {
        PaperTrailErrorEvent: emitter
    }
});

jest.mock("../handlers", () => ({
    logAndNotify: jest.fn(),
    handleError: jest.fn(),
}));

jest.mock("../middleware", () => ({
    expressLogErrors: jest.fn(),
    expressHandleErrors: jest.fn()
}));

const logger = require("@firstfleet/fflogger");
const { slackChannels, errorLevels } = require("../constants");
const handlers = require("../handlers");
const { expressLogErrors, expressHandleErrors } = require("../middleware");
const main = require("../index");

const appName = process.env.PAPERTRAIL_PROGRAM;

jest.useFakeTimers();

beforeEach(() => {
    jest.resetModules();
});

afterEach(() => {
    jest.restoreAllMocks();
});

describe("index.js", () => {
    const testErrorMessage = "test message";
    const testError = new Error(testErrorMessage);

    it("Should handle papertrail errors by calling console.error and logAndNotify emmited from logger.PaperTrailErrorEvent", () => {
        const consoleErrorSpy = jest.spyOn(console, 'error').mockImplementation()

        setTimeout(() => {
            logger.PaperTrailErrorEvent.emit("error", testError)
        }, 100);
        jest.advanceTimersByTime(100)

        expect(consoleErrorSpy).toHaveBeenCalledWith(testError);
        expect(handlers.logAndNotify).toHaveBeenCalledWith(
            appName,
            "Error Handler Papertrail Init",
            `Unable to log to Papertrail: ${testError.message}`,
            testError.stack,
            slackChannels.WARNING
        );

        consoleErrorSpy.mockClear();
    });

    it("Should call console.error if an error is thrown in PaperTrailErrorEvent.on('error')", () => {
        const consoleErrorSpy = jest.spyOn(console, 'error').mockImplementation()

        const logError = new Error("log error");

        handlers.logAndNotify.mockImplementation(() => { throw logError })

        setTimeout(() => {
            logger.PaperTrailErrorEvent.emit("error", testError)
        }, 100);
        jest.advanceTimersByTime(100)

        expect(consoleErrorSpy).toHaveBeenNthCalledWith(2, logError);

        consoleErrorSpy.mockClear();
        handlers.logAndNotify.mockReset();
    });

    it("Should call console.error and logAndNotify, then gracefully shutdown the server if there is an uncaught exception", () => {
        const consoleErrorSpy = jest.spyOn(console, 'error').mockImplementation()
        const processExitSpy = jest.spyOn(process, 'exit').mockImplementation();

        setTimeout(() => {
            process.emit("uncaughtException", testError)
        }, 100);
        jest.advanceTimersByTime(100)

        expect(consoleErrorSpy).toHaveBeenCalledWith(testError);
        expect(handlers.logAndNotify).toHaveBeenCalledWith(
            appName,
            'GLOBAL - uncaughtException',
            testError.message,
            testError.stack,
            slackChannels.CRITICAL
        );

        jest.advanceTimersByTime(1000);
        expect(processExitSpy).toHaveBeenCalledWith(1);

        consoleErrorSpy.mockClear()
        processExitSpy.mockClear()
    });

    it("Should call console.error and then gracefully shutdown the server if there is an uncaught exception and logAndNotify throws an error", () => {
        const consoleErrorSpy = jest.spyOn(console, 'error').mockImplementation()
        const processExitSpy = jest.spyOn(process, 'exit').mockImplementation();

        const logError = new Error("log error");

        handlers.logAndNotify.mockImplementation(() => { throw logError })

        setTimeout(() => {
            process.emit("uncaughtException", testError)
        }, 100);
        jest.advanceTimersByTime(100)

        expect(consoleErrorSpy).toHaveBeenNthCalledWith(2, logError);
        expect(processExitSpy).toHaveBeenCalledWith(1);

        consoleErrorSpy.mockClear()
        processExitSpy.mockClear()
        handlers.logAndNotify.mockReset();
    });

    it("Should call console.error and logAndNotify if there is an unhandledRejection error thrown", () => {
        const consoleErrorSpy = jest.spyOn(console, 'error').mockImplementation()

        setTimeout(() => {
            process.emit("unhandledRejection", testError)
        }, 100);
        jest.advanceTimersByTime(100)

        expect(consoleErrorSpy).toHaveBeenCalledWith(testError);
        expect(handlers.logAndNotify).toHaveBeenCalledWith(
            appName,
            'GLOBAL - unhandledRejection',
            testError.message,
            testError.stack,
            slackChannels.CRITICAL
        );

        consoleErrorSpy.mockClear()
    });

    it("Should call console.error if there is an unhandledRejection error thrown, and logAndNotify throws an error", () => {
        const consoleErrorSpy = jest.spyOn(console, 'error').mockImplementation()

        const logError = new Error("log error");

        handlers.logAndNotify.mockImplementation(() => { throw logError })

        setTimeout(() => {
            process.emit("unhandledRejection", testError)
        }, 100);
        jest.advanceTimersByTime(100)

        expect(consoleErrorSpy).toHaveBeenNthCalledWith(2, logError);

        consoleErrorSpy.mockClear()
        handlers.logAndNotify.mockReset();
    });
});


